# 23 "lexer.mll"
 
open Lexing
open Ast

type error_msg =
	| Invalid_character of char
	| Unterminated_string
	| Unterminated_regexp
	| Unclosed_comment
	| Unclosed_code
	| Invalid_escape
	| Invalid_option

exception Error of error_msg * pos

let error_msg = function
	| Invalid_character c when int_of_char c > 32 && int_of_char c < 128 -> Printf.sprintf "Invalid character '%c'" c
	| Invalid_character c -> Printf.sprintf "Invalid character 0x%.2X" (int_of_char c)
	| Unterminated_string -> "Unterminated string"
	| Unterminated_regexp -> "Unterminated regular expression"
	| Unclosed_comment -> "Unclosed comment"
	| Unclosed_code -> "Unclosed code string"
	| Invalid_escape -> "Invalid escape sequence"
	| Invalid_option -> "Invalid regular expression option"

type lexer_file = {
	lfile : string;
	mutable lline : int;
	mutable lmaxline : int;
	mutable llines : (int * int) list;
	mutable lalines : (int * int) array;
	mutable lstrings : int list;
}

let make_file file =
	{
		lfile = file;
		lline = 1;
		lmaxline = 1;
		llines = [0,1];
		lalines = [|0,1|];
		lstrings = [];
	}


let cur = ref (make_file "")

let all_files = Hashtbl.create 0

let buf = Buffer.create 100

let error e pos =
	raise (Error (e,{ pmin = pos; pmax = pos; pfile = !cur.lfile }))

let keywords =
	let h = Hashtbl.create 3 in
	List.iter (fun k -> Hashtbl.add h (s_keyword k) k)
		[Function;Class;Static;Var;If;Else;While;Do;For;
		Break;Return;Continue;Extends;Implements;Import;
		Switch;Case;Default;Public;Private;Try;Untyped;
		Catch;New;This;Throw;Extern;Enum;In;Interface;
		Cast;Override;Dynamic;Typedef;Package;
		Inline;Using;Null;True;False;Abstract;Macro];
	h

let init file do_add =
	let f = make_file file in
	cur := f;
	if do_add then Hashtbl.replace all_files file f

let save() =
	!cur

let restore c =
	cur := c

let newline lexbuf =
	let cur = !cur in
	cur.lline <- cur.lline + 1;
	cur.llines <- (lexeme_end lexbuf,cur.lline) :: cur.llines

let fmt_pos p =
	p.pmin + (p.pmax - p.pmin) * 1000000

let add_fmt_string p =
	let file = (try
		Hashtbl.find all_files p.pfile
	with Not_found ->
		let f = make_file p.pfile in
		Hashtbl.replace all_files p.pfile f;
		f
	) in
	file.lstrings <- (fmt_pos p) :: file.lstrings

let fast_add_fmt_string p =
	let cur = !cur in
	cur.lstrings <- (fmt_pos p) :: cur.lstrings

let is_fmt_string p =
	try
		let file = Hashtbl.find all_files p.pfile in
		List.mem (fmt_pos p) file.lstrings
	with Not_found ->
		false

let remove_fmt_string p =
	try
		let file = Hashtbl.find all_files p.pfile in
		file.lstrings <- List.filter ((<>) (fmt_pos p)) file.lstrings
	with Not_found ->
		()

let find_line p f =
	(* rebuild cache if we have a new line *)
	if f.lmaxline <> f.lline then begin
		f.lmaxline <- f.lline;
		f.lalines <- Array.of_list (List.rev f.llines);
	end;
	let rec loop min max =
		let med = (min + max) lsr 1 in
		let lp, line = Array.unsafe_get f.lalines med in
		if med = min then
			line, p - lp
		else if lp > p then
			loop min med
		else
			loop med max
	in
	loop 0 (Array.length f.lalines)

(* resolve a position within a non-haxe file by counting newlines *)
let resolve_pos file =
	let ch = open_in_bin file in
	let f = make_file file in
	let rec loop p =
		let inc i () =
			f.lline <- f.lline + 1;
			f.llines <- (p + i,f.lline) :: f.llines;
			i
		in
		let i = match input_char ch with
			| '\n' -> inc 1
			| '\r' ->
				ignore(input_char ch);
				inc 2
			| _ -> fun () -> 1
		in
		loop (p + i())
	in
	try
		loop 0
	with End_of_file ->
		close_in ch;
		f

let find_file file =
	try Hashtbl.find all_files file with Not_found -> try resolve_pos file with Sys_error _ -> make_file file

let find_pos p =
	find_line p.pmin (find_file p.pfile)

let get_error_line p =
	let l, _ = find_pos p in
	l

let get_error_pos printer p =
	if p.pmin = -1 then
		"(unknown)"
	else
		let file = find_file p.pfile in
		let l1, p1 = find_line p.pmin file in
		let l2, p2 = find_line p.pmax file in
		if l1 = l2 then begin
			let s = (if p1 = p2 then Printf.sprintf " %d" p1 else Printf.sprintf "s %d-%d" p1 p2) in
			Printf.sprintf "%s character%s" (printer p.pfile l1) s
		end else
			Printf.sprintf "%s lines %d-%d" (printer p.pfile l1) l1 l2

let reset() = Buffer.reset buf
let contents() = Buffer.contents buf
let store lexbuf = Buffer.add_string buf (lexeme lexbuf)
let add c = Buffer.add_string buf c

let mk_tok t pmin pmax =
	t , { pfile = !cur.lfile; pmin = pmin; pmax = pmax }

let mk lexbuf t =
	mk_tok t (lexeme_start lexbuf) (lexeme_end lexbuf)

let mk_ident lexbuf =
	let s = lexeme lexbuf in
	mk lexbuf (try Kwd (Hashtbl.find keywords s) with Not_found -> Const (Ident s))

let invalid_char lexbuf =
	error (Invalid_character (lexeme_char lexbuf 0)) (lexeme_start lexbuf)


# 200 "lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\253\255\000\000\000\000\000\000\255\255\001\000\006\000\
    \190\255\095\000\198\000\017\001\092\001\120\001\196\255\197\255\
    \199\255\200\255\201\255\202\255\203\255\204\255\205\255\206\255\
    \217\255\218\255\219\255\220\255\037\000\072\000\075\000\041\000\
    \076\000\094\000\181\000\077\000\092\000\095\000\098\000\179\000\
    \179\000\197\001\198\001\252\255\000\000\000\000\255\255\253\255\
    \220\001\232\001\254\001\036\002\096\000\064\002\074\002\096\002\
    \106\002\245\255\116\002\128\002\107\000\138\002\224\255\198\255\
    \233\255\023\001\236\255\243\255\225\255\235\255\242\255\195\255\
    \240\255\228\255\239\255\227\255\238\255\237\255\234\255\229\255\
    \093\000\232\255\223\255\231\255\230\255\148\002\223\002\042\003\
    \117\003\141\001\142\001\111\000\002\000\254\255\255\255\253\255\
    \047\002\048\002\250\255\152\000\003\000\254\255\255\255\252\255\
    \253\255\230\003\231\003\209\000\250\255\202\000\007\000\254\255\
    \255\255\249\255\251\255\252\255\248\255\237\003\238\003\250\255\
    \251\255\252\255\033\001\253\255\008\000\254\255\255\255\025\002\
    \249\255\166\003\198\003\246\255\002\004\255\255\247\255\080\004\
    \249\255\250\255\251\255\252\255\253\255\254\255\103\004\194\004\
    \217\004\248\255\223\004\254\255\255\255";
  Lexing.lex_backtrk = 
   "\002\000\255\255\255\255\255\255\255\255\255\255\001\000\255\255\
    \255\255\064\000\063\000\063\000\062\000\065\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\033\000\048\000\034\000\045\000\
    \043\000\042\000\041\000\040\000\014\000\047\000\044\000\046\000\
    \039\000\005\000\005\000\255\255\003\000\001\000\255\255\255\255\
    \255\255\255\255\255\255\004\000\255\255\255\255\006\000\009\000\
    \255\255\255\255\008\000\255\255\255\255\007\000\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \029\000\255\255\255\255\255\255\255\255\061\000\061\000\061\000\
    \063\000\255\255\004\000\003\000\001\000\255\255\255\255\255\255\
    \255\255\006\000\255\255\004\000\001\000\255\255\255\255\255\255\
    \255\255\255\255\008\000\006\000\255\255\002\000\001\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\002\000\255\255\001\000\255\255\255\255\007\000\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\008\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\002\000\255\255\255\255";
  Lexing.lex_default = 
   "\255\255\000\000\255\255\255\255\255\255\000\000\006\000\008\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\065\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\090\000\090\000\255\255\255\255\000\000\000\000\000\000\
    \097\000\097\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\106\000\106\000\255\255\000\000\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\118\000\118\000\000\000\
    \000\000\000\000\255\255\000\000\255\255\000\000\000\000\127\000\
    \000\000\130\000\130\000\000\000\134\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\000\000\255\255\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\045\000\047\000\255\255\093\000\101\000\255\255\045\000\
    \043\000\111\000\125\000\044\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \045\000\006\000\000\000\002\000\000\000\000\000\045\000\028\000\
    \015\000\013\000\012\000\035\000\034\000\014\000\019\000\018\000\
    \031\000\038\000\024\000\037\000\040\000\039\000\042\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \025\000\026\000\030\000\029\000\027\000\017\000\016\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\023\000\084\000\022\000\032\000\011\000\078\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\021\000\033\000\020\000\036\000\083\000\082\000\080\000\
    \079\000\077\000\072\000\071\000\070\000\067\000\057\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\062\000\081\000\076\000\069\000\068\000\095\000\066\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\104\000\004\000\000\000\000\000\009\000\005\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\075\000\073\000\000\000\063\000\000\000\000\000\
    \000\000\060\000\065\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\000\000\113\000\003\000\
    \064\000\114\000\074\000\000\000\103\000\113\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \001\000\255\255\000\000\000\000\000\000\000\000\046\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\255\255\000\000\000\000\255\255\010\000\115\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\128\000\116\000\000\000\000\000\000\000\
    \127\000\000\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\000\000\000\000\000\000\000\000\
    \011\000\000\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\000\000\093\000\
    \255\255\000\000\092\000\255\255\000\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\091\000\
    \255\255\000\000\000\000\012\000\000\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\086\000\
    \000\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\049\000\049\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\059\000\
    \000\000\059\000\048\000\048\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\052\000\255\255\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\000\000\255\255\000\000\000\000\255\255\000\000\
    \000\000\000\000\048\000\048\000\000\000\053\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \000\000\101\000\255\255\000\000\100\000\255\255\050\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\053\000\000\000\000\000\
    \000\000\098\000\255\255\000\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\000\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\056\000\000\000\056\000\000\000\000\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\000\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\099\000\255\255\094\000\255\255\053\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\000\000\053\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\000\000\
    \000\000\000\000\000\000\085\000\000\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\102\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\086\000\000\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\000\000\000\000\000\000\
    \000\000\087\000\000\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\000\000\
    \133\000\000\000\000\000\133\000\000\000\000\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \255\255\000\000\000\000\255\255\088\000\131\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \111\000\255\255\000\000\110\000\255\255\255\255\000\000\125\000\
    \255\255\000\000\124\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\132\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\107\000\255\255\000\000\108\000\255\255\120\000\
    \255\255\000\000\000\000\000\000\119\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\122\000\255\255\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\137\000\000\000\
    \000\000\000\000\137\000\137\000\137\000\137\000\000\000\137\000\
    \137\000\141\000\137\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\000\000\000\000\000\000\000\000\
    \000\000\137\000\109\000\255\255\136\000\000\000\136\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\136\000\000\000\135\000\
    \000\000\136\000\000\000\000\000\000\000\137\000\137\000\137\000\
    \137\000\000\000\000\000\000\000\136\000\000\000\136\000\000\000\
    \123\000\255\255\121\000\255\255\000\000\000\000\000\000\000\000\
    \139\000\000\000\000\000\000\000\140\000\136\000\138\000\135\000\
    \000\000\136\000\136\000\000\000\000\000\137\000\137\000\137\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\142\000\142\000\142\000\142\000\142\000\142\000\143\000\
    \143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\
    \143\000\000\000\000\000\000\000\000\000\000\000\133\000\000\000\
    \143\000\143\000\143\000\143\000\143\000\143\000\000\000\000\000\
    \000\000\142\000\142\000\142\000\142\000\142\000\142\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \143\000\143\000\143\000\143\000\143\000\143\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\112\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\126\000\255\255\000\000\
    \000\000\000\000\144\000\144\000\144\000\144\000\144\000\144\000\
    \144\000\144\000\144\000\144\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\144\000\144\000\144\000\144\000\144\000\
    \144\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \000\000\000\000\000\000\144\000\144\000\144\000\144\000\144\000\
    \144\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \147\000\147\000\147\000\147\000\147\000\147\000\148\000\147\000\
    \148\000\147\000\147\000\147\000\148\000\147\000\147\000\147\000\
    \147\000\147\000\148\000\147\000\148\000\147\000\147\000\147\000\
    \147\000\147\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\045\000\044\000\006\000\092\000\100\000\006\000\007\000\
    \007\000\110\000\124\000\007\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \045\000\002\000\255\255\000\000\255\255\255\255\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\028\000\007\000\007\000\007\000\031\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\029\000\029\000\030\000\
    \030\000\032\000\035\000\036\000\037\000\038\000\052\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\060\000\080\000\033\000\037\000\037\000\091\000\038\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\099\000\003\000\255\255\255\255\009\000\004\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\
    \009\000\009\000\033\000\034\000\255\255\039\000\255\255\255\255\
    \255\255\040\000\039\000\040\000\040\000\040\000\040\000\040\000\
    \040\000\040\000\040\000\040\000\040\000\255\255\109\000\000\000\
    \039\000\109\000\034\000\255\255\099\000\107\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \000\000\006\000\255\255\255\255\255\255\255\255\007\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\065\000\255\255\255\255\065\000\010\000\109\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\122\000\107\000\255\255\255\255\255\255\
    \122\000\255\255\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\255\255\255\255\255\255\255\255\
    \011\000\255\255\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\
    \011\000\011\000\011\000\011\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\255\255\089\000\
    \090\000\255\255\089\000\090\000\255\255\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\089\000\
    \090\000\255\255\255\255\012\000\255\255\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\013\000\
    \255\255\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\041\000\042\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\048\000\
    \255\255\048\000\041\000\042\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\049\000\065\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\255\255\127\000\255\255\255\255\127\000\255\255\
    \255\255\255\255\041\000\042\000\255\255\049\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \255\255\096\000\097\000\255\255\096\000\097\000\042\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\049\000\255\255\255\255\
    \255\255\096\000\097\000\255\255\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\255\255\050\000\
    \050\000\050\000\050\000\050\000\050\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\053\000\255\255\053\000\255\255\255\255\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\255\255\051\000\051\000\051\000\
    \051\000\051\000\051\000\096\000\097\000\089\000\090\000\054\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\255\255\054\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\255\255\
    \255\255\255\255\255\255\085\000\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\127\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\096\000\
    \097\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\086\000\255\255\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\255\255\255\255\255\255\
    \255\255\087\000\255\255\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\255\255\
    \129\000\255\255\255\255\129\000\255\255\255\255\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \130\000\255\255\255\255\130\000\088\000\129\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \105\000\106\000\255\255\105\000\106\000\130\000\255\255\117\000\
    \118\000\255\255\117\000\118\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\129\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\105\000\106\000\255\255\105\000\106\000\117\000\
    \118\000\255\255\255\255\255\255\117\000\118\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\117\000\118\000\255\255\255\255\
    \255\255\255\255\130\000\255\255\255\255\255\255\132\000\255\255\
    \255\255\255\255\132\000\132\000\132\000\132\000\255\255\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\255\255\255\255\255\255\255\255\
    \255\255\132\000\105\000\106\000\132\000\255\255\132\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\132\000\255\255\132\000\
    \255\255\132\000\255\255\255\255\255\255\132\000\132\000\132\000\
    \132\000\255\255\255\255\255\255\132\000\255\255\132\000\255\255\
    \117\000\118\000\117\000\118\000\255\255\255\255\255\255\255\255\
    \132\000\255\255\255\255\255\255\132\000\132\000\132\000\132\000\
    \255\255\132\000\132\000\255\255\255\255\132\000\132\000\132\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\135\000\135\000\135\000\135\000\135\000\135\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\255\255\255\255\255\255\255\255\255\255\129\000\255\255\
    \142\000\142\000\142\000\142\000\142\000\142\000\255\255\255\255\
    \255\255\135\000\135\000\135\000\135\000\135\000\135\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\130\000\255\255\
    \142\000\142\000\142\000\142\000\142\000\142\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\105\000\106\000\
    \255\255\255\255\255\255\255\255\255\255\117\000\118\000\255\255\
    \255\255\255\255\143\000\143\000\143\000\143\000\143\000\143\000\
    \143\000\143\000\143\000\143\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\132\000\143\000\143\000\143\000\143\000\143\000\
    \143\000\144\000\144\000\144\000\144\000\144\000\144\000\144\000\
    \144\000\144\000\144\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\144\000\144\000\144\000\144\000\144\000\144\000\
    \255\255\255\255\255\255\143\000\143\000\143\000\143\000\143\000\
    \143\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\144\000\144\000\144\000\144\000\144\000\144\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    ";
  Lexing.lex_base_code = 
   "";
  Lexing.lex_backtrk_code = 
   "";
  Lexing.lex_default_code = 
   "";
  Lexing.lex_trans_code = 
   "";
  Lexing.lex_check_code = 
   "";
  Lexing.lex_code = 
   "";
}

let rec skip_header lexbuf =
    __ocaml_lex_skip_header_rec lexbuf 0
and __ocaml_lex_skip_header_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 226 "lexer.mll"
                  ( skip_header lexbuf )
# 663 "lexer.ml"

  | 1 ->
# 227 "lexer.mll"
                      ( skip_header lexbuf )
# 668 "lexer.ml"

  | 2 ->
# 228 "lexer.mll"
            ( )
# 673 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_skip_header_rec lexbuf __ocaml_lex_state

and token lexbuf =
    __ocaml_lex_token_rec lexbuf 7
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 231 "lexer.mll"
       ( mk lexbuf Eof )
# 685 "lexer.ml"

  | 1 ->
# 232 "lexer.mll"
               ( token lexbuf )
# 690 "lexer.ml"

  | 2 ->
# 233 "lexer.mll"
          ( newline lexbuf; token lexbuf )
# 695 "lexer.ml"

  | 3 ->
# 234 "lexer.mll"
               ( newline lexbuf; token lexbuf )
# 700 "lexer.ml"

  | 4 ->
# 235 "lexer.mll"
                                   ( mk lexbuf (Const (Int (lexeme lexbuf))) )
# 705 "lexer.ml"

  | 5 ->
# 236 "lexer.mll"
           ( mk lexbuf (Const (Int (lexeme lexbuf))) )
# 710 "lexer.ml"

  | 6 ->
# 237 "lexer.mll"
                          ( mk lexbuf (Const (Float (lexeme lexbuf))) )
# 715 "lexer.ml"

  | 7 ->
# 238 "lexer.mll"
                  ( mk lexbuf (Const (Float (lexeme lexbuf))) )
# 720 "lexer.ml"

  | 8 ->
# 239 "lexer.mll"
                                           ( mk lexbuf (Const (Float (lexeme lexbuf))) )
# 725 "lexer.ml"

  | 9 ->
# 240 "lexer.mll"
                                                          ( mk lexbuf (Const (Float (lexeme lexbuf))) )
# 730 "lexer.ml"

  | 10 ->
# 241 "lexer.mll"
                 (
			let s = lexeme lexbuf in
			mk lexbuf (IntInterval (String.sub s 0 (String.length s - 3)))
		)
# 738 "lexer.ml"

  | 11 ->
# 245 "lexer.mll"
                       (
			let s = lexeme lexbuf in
			mk lexbuf (CommentLine (String.sub s 2 ((String.length s)-2)))
		)
# 746 "lexer.ml"

  | 12 ->
# 249 "lexer.mll"
        ( mk lexbuf (Unop Increment) )
# 751 "lexer.ml"

  | 13 ->
# 250 "lexer.mll"
        ( mk lexbuf (Unop Decrement) )
# 756 "lexer.ml"

  | 14 ->
# 251 "lexer.mll"
        ( mk lexbuf (Unop NegBits) )
# 761 "lexer.ml"

  | 15 ->
# 252 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpMod)) )
# 766 "lexer.ml"

  | 16 ->
# 253 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpAnd)) )
# 771 "lexer.ml"

  | 17 ->
# 254 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpOr)) )
# 776 "lexer.ml"

  | 18 ->
# 255 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpXor)) )
# 781 "lexer.ml"

  | 19 ->
# 256 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpAdd)) )
# 786 "lexer.ml"

  | 20 ->
# 257 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpSub)) )
# 791 "lexer.ml"

  | 21 ->
# 258 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpMult)) )
# 796 "lexer.ml"

  | 22 ->
# 259 "lexer.mll"
        ( mk lexbuf (Binop (OpAssignOp OpDiv)) )
# 801 "lexer.ml"

  | 23 ->
# 260 "lexer.mll"
         ( mk lexbuf (Binop (OpAssignOp OpShl)) )
# 806 "lexer.ml"

  | 24 ->
# 263 "lexer.mll"
        ( mk lexbuf (Binop OpEq) )
# 811 "lexer.ml"

  | 25 ->
# 264 "lexer.mll"
        ( mk lexbuf (Binop OpNotEq) )
# 816 "lexer.ml"

  | 26 ->
# 265 "lexer.mll"
        ( mk lexbuf (Binop OpLte) )
# 821 "lexer.ml"

  | 27 ->
# 267 "lexer.mll"
        ( mk lexbuf (Binop OpBoolAnd) )
# 826 "lexer.ml"

  | 28 ->
# 268 "lexer.mll"
        ( mk lexbuf (Binop OpBoolOr) )
# 831 "lexer.ml"

  | 29 ->
# 269 "lexer.mll"
        ( mk lexbuf (Binop OpShl) )
# 836 "lexer.ml"

  | 30 ->
# 270 "lexer.mll"
        ( mk lexbuf Arrow )
# 841 "lexer.ml"

  | 31 ->
# 271 "lexer.mll"
         ( mk lexbuf (Binop OpInterval) )
# 846 "lexer.ml"

  | 32 ->
# 272 "lexer.mll"
        ( mk lexbuf (Binop OpArrow))
# 851 "lexer.ml"

  | 33 ->
# 273 "lexer.mll"
       ( mk lexbuf (Unop Not) )
# 856 "lexer.ml"

  | 34 ->
# 274 "lexer.mll"
       ( mk lexbuf (Binop OpLt) )
# 861 "lexer.ml"

  | 35 ->
# 275 "lexer.mll"
       ( mk lexbuf (Binop OpGt) )
# 866 "lexer.ml"

  | 36 ->
# 276 "lexer.mll"
       ( mk lexbuf Semicolon )
# 871 "lexer.ml"

  | 37 ->
# 277 "lexer.mll"
       ( mk lexbuf DblDot )
# 876 "lexer.ml"

  | 38 ->
# 278 "lexer.mll"
       ( mk lexbuf Comma )
# 881 "lexer.ml"

  | 39 ->
# 279 "lexer.mll"
       ( mk lexbuf Dot )
# 886 "lexer.ml"

  | 40 ->
# 280 "lexer.mll"
       ( mk lexbuf (Binop OpMod) )
# 891 "lexer.ml"

  | 41 ->
# 281 "lexer.mll"
       ( mk lexbuf (Binop OpAnd) )
# 896 "lexer.ml"

  | 42 ->
# 282 "lexer.mll"
       ( mk lexbuf (Binop OpOr) )
# 901 "lexer.ml"

  | 43 ->
# 283 "lexer.mll"
       ( mk lexbuf (Binop OpXor) )
# 906 "lexer.ml"

  | 44 ->
# 284 "lexer.mll"
       ( mk lexbuf (Binop OpAdd) )
# 911 "lexer.ml"

  | 45 ->
# 285 "lexer.mll"
       ( mk lexbuf (Binop OpMult) )
# 916 "lexer.ml"

  | 46 ->
# 286 "lexer.mll"
       ( mk lexbuf (Binop OpDiv) )
# 921 "lexer.ml"

  | 47 ->
# 287 "lexer.mll"
       ( mk lexbuf (Binop OpSub) )
# 926 "lexer.ml"

  | 48 ->
# 288 "lexer.mll"
       ( mk lexbuf (Binop OpAssign) )
# 931 "lexer.ml"

  | 49 ->
# 289 "lexer.mll"
       ( mk lexbuf BkOpen )
# 936 "lexer.ml"

  | 50 ->
# 290 "lexer.mll"
       ( mk lexbuf BkClose )
# 941 "lexer.ml"

  | 51 ->
# 291 "lexer.mll"
       ( mk lexbuf BrOpen )
# 946 "lexer.ml"

  | 52 ->
# 292 "lexer.mll"
       ( mk lexbuf BrClose )
# 951 "lexer.ml"

  | 53 ->
# 293 "lexer.mll"
       ( mk lexbuf POpen )
# 956 "lexer.ml"

  | 54 ->
# 294 "lexer.mll"
       ( mk lexbuf PClose )
# 961 "lexer.ml"

  | 55 ->
# 295 "lexer.mll"
       ( mk lexbuf Question )
# 966 "lexer.ml"

  | 56 ->
# 296 "lexer.mll"
       ( mk lexbuf At )
# 971 "lexer.ml"

  | 57 ->
# 297 "lexer.mll"
        (
			reset();
			let pmin = lexeme_start lexbuf in
			let pmax = (try comment lexbuf with Exit -> error Unclosed_comment pmin) in
			mk_tok (Comment (contents())) pmin pmax;
		)
# 981 "lexer.ml"

  | 58 ->
# 303 "lexer.mll"
       (
			reset();
			let pmin = lexeme_start lexbuf in
			let pmax = (try string lexbuf with Exit -> error Unterminated_string pmin) in
			let str = (try unescape (contents()) with Exit -> error Invalid_escape pmin) in
			mk_tok (Const (String str)) pmin pmax;
		)
# 992 "lexer.ml"

  | 59 ->
# 310 "lexer.mll"
       (
			reset();
			let pmin = lexeme_start lexbuf in
			let pmax = (try string2 lexbuf with Exit -> error Unterminated_string pmin) in
			let str = (try unescape (contents()) with Exit -> error Invalid_escape pmin) in
			let t = mk_tok (Const (String str)) pmin pmax in
			fast_add_fmt_string (snd t);
			t
		)
# 1005 "lexer.ml"

  | 60 ->
# 319 "lexer.mll"
        (
			reset();
			let pmin = lexeme_start lexbuf in
			let options, pmax = (try regexp lexbuf with Exit -> error Unterminated_regexp pmin) in
			let str = contents() in
			mk_tok (Const (Regexp (str,options))) pmin pmax;
		)
# 1016 "lexer.ml"

  | 61 ->
# 326 "lexer.mll"
             (
			let v = lexeme lexbuf in
			let v = String.sub v 1 (String.length v - 1) in
			mk lexbuf (Sharp v)
		)
# 1025 "lexer.ml"

  | 62 ->
# 331 "lexer.mll"
                                      (
			let v = lexeme lexbuf in
			let v = String.sub v 1 (String.length v - 1) in
			mk lexbuf (Dollar v)
		)
# 1034 "lexer.ml"

  | 63 ->
# 336 "lexer.mll"
         ( mk_ident lexbuf )
# 1039 "lexer.ml"

  | 64 ->
# 337 "lexer.mll"
          ( mk lexbuf (Const (Ident (lexeme lexbuf))) )
# 1044 "lexer.ml"

  | 65 ->
# 338 "lexer.mll"
     ( invalid_char lexbuf )
# 1049 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and comment lexbuf =
    __ocaml_lex_comment_rec lexbuf 89
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 341 "lexer.mll"
       ( raise Exit )
# 1061 "lexer.ml"

  | 1 ->
# 342 "lexer.mll"
                        ( newline lexbuf; store lexbuf; comment lexbuf )
# 1066 "lexer.ml"

  | 2 ->
# 343 "lexer.mll"
        ( lexeme_end lexbuf )
# 1071 "lexer.ml"

  | 3 ->
# 344 "lexer.mll"
       ( store lexbuf; comment lexbuf )
# 1076 "lexer.ml"

  | 4 ->
# 345 "lexer.mll"
                     ( store lexbuf; comment lexbuf )
# 1081 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and string lexbuf =
    __ocaml_lex_string_rec lexbuf 96
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 348 "lexer.mll"
       ( raise Exit )
# 1093 "lexer.ml"

  | 1 ->
# 349 "lexer.mll"
                        ( newline lexbuf; store lexbuf; string lexbuf )
# 1098 "lexer.ml"

  | 2 ->
# 350 "lexer.mll"
          ( store lexbuf; string lexbuf )
# 1103 "lexer.ml"

  | 3 ->
# 351 "lexer.mll"
          ( store lexbuf; string lexbuf )
# 1108 "lexer.ml"

  | 4 ->
# 352 "lexer.mll"
        ( store lexbuf; string lexbuf )
# 1113 "lexer.ml"

  | 5 ->
# 353 "lexer.mll"
       ( lexeme_end lexbuf )
# 1118 "lexer.ml"

  | 6 ->
# 354 "lexer.mll"
                          ( store lexbuf; string lexbuf )
# 1123 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and string2 lexbuf =
    __ocaml_lex_string2_rec lexbuf 105
and __ocaml_lex_string2_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 357 "lexer.mll"
       ( raise Exit )
# 1135 "lexer.ml"

  | 1 ->
# 358 "lexer.mll"
                        ( newline lexbuf; store lexbuf; string2 lexbuf )
# 1140 "lexer.ml"

  | 2 ->
# 359 "lexer.mll"
        ( store lexbuf; string2 lexbuf )
# 1145 "lexer.ml"

  | 3 ->
# 360 "lexer.mll"
          ( store lexbuf; string2 lexbuf )
# 1150 "lexer.ml"

  | 4 ->
# 361 "lexer.mll"
         ( store lexbuf; string2 lexbuf )
# 1155 "lexer.ml"

  | 5 ->
# 362 "lexer.mll"
       ( lexeme_end lexbuf )
# 1160 "lexer.ml"

  | 6 ->
# 363 "lexer.mll"
                      ( store lexbuf; string2 lexbuf )
# 1165 "lexer.ml"

  | 7 ->
# 364 "lexer.mll"
        (
		let pmin = lexeme_start lexbuf in
		store lexbuf;
		(try code_string lexbuf with Exit -> error Unclosed_code pmin);
		string2 lexbuf;
	)
# 1175 "lexer.ml"

  | 8 ->
# 370 "lexer.mll"
                               ( store lexbuf; string2 lexbuf )
# 1180 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string2_rec lexbuf __ocaml_lex_state

and code_string lexbuf =
    __ocaml_lex_code_string_rec lexbuf 117
and __ocaml_lex_code_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 373 "lexer.mll"
       ( raise Exit )
# 1192 "lexer.ml"

  | 1 ->
# 374 "lexer.mll"
                        ( newline lexbuf; store lexbuf; code_string lexbuf )
# 1197 "lexer.ml"

  | 2 ->
# 375 "lexer.mll"
             ( store lexbuf; code_string lexbuf )
# 1202 "lexer.ml"

  | 3 ->
# 376 "lexer.mll"
       ( store lexbuf; (* stop *) )
# 1207 "lexer.ml"

  | 4 ->
# 377 "lexer.mll"
       (
		add "\"";
		let pmin = lexeme_start lexbuf in
		(try ignore(string lexbuf) with Exit -> error Unterminated_string pmin);
		add "\"";
		code_string lexbuf;
	)
# 1218 "lexer.ml"

  | 5 ->
# 384 "lexer.mll"
       (
		add "'";
		let pmin = lexeme_start lexbuf in
		let pmax = (try string2 lexbuf with Exit -> error Unterminated_string pmin) in
		add "'";
		fast_add_fmt_string { pfile = !cur.lfile; pmin = pmin; pmax = pmax };
		code_string lexbuf;
	)
# 1230 "lexer.ml"

  | 6 ->
# 392 "lexer.mll"
        (
		let pmin = lexeme_start lexbuf in
		(try ignore(comment lexbuf) with Exit -> error Unclosed_comment pmin);
		code_string lexbuf;
	)
# 1239 "lexer.ml"

  | 7 ->
# 397 "lexer.mll"
                      ( store lexbuf; code_string lexbuf; )
# 1244 "lexer.ml"

  | 8 ->
# 398 "lexer.mll"
                                       ( store lexbuf; code_string lexbuf; )
# 1249 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_code_string_rec lexbuf __ocaml_lex_state

and regexp lexbuf =
    __ocaml_lex_regexp_rec lexbuf 129
and __ocaml_lex_regexp_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 401 "lexer.mll"
                     ( raise Exit )
# 1261 "lexer.ml"

  | 1 ->
# 402 "lexer.mll"
            ( add "/"; regexp lexbuf )
# 1266 "lexer.ml"

  | 2 ->
# 403 "lexer.mll"
            ( add "\r"; regexp lexbuf )
# 1271 "lexer.ml"

  | 3 ->
# 404 "lexer.mll"
            ( add "\n"; regexp lexbuf )
# 1276 "lexer.ml"

  | 4 ->
# 405 "lexer.mll"
            ( add "\t"; regexp lexbuf )
# 1281 "lexer.ml"

  | 5 ->
# 406 "lexer.mll"
                                                                               ( add (lexeme lexbuf); regexp lexbuf )
# 1286 "lexer.ml"

  | 6 ->
# 407 "lexer.mll"
                                              ( add (lexeme lexbuf); regexp lexbuf )
# 1291 "lexer.ml"

  | 7 ->
# 408 "lexer.mll"
                                                                                                                          ( add (lexeme lexbuf); regexp lexbuf )
# 1296 "lexer.ml"

  | 8 ->
# 409 "lexer.mll"
                 ( error (Invalid_character (lexeme lexbuf).[1]) (lexeme_end lexbuf - 1) )
# 1301 "lexer.ml"

  | 9 ->
# 410 "lexer.mll"
       ( regexp_options lexbuf, lexeme_end lexbuf )
# 1306 "lexer.ml"

  | 10 ->
# 411 "lexer.mll"
                           ( store lexbuf; regexp lexbuf )
# 1311 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_regexp_rec lexbuf __ocaml_lex_state

and regexp_options lexbuf =
    __ocaml_lex_regexp_options_rec lexbuf 146
and __ocaml_lex_regexp_options_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 414 "lexer.mll"
                               (
			let l = lexeme lexbuf in
			l ^ regexp_options lexbuf
		)
# 1326 "lexer.ml"

  | 1 ->
# 418 "lexer.mll"
               ( error Invalid_option (lexeme_start lexbuf) )
# 1331 "lexer.ml"

  | 2 ->
# 419 "lexer.mll"
      ( "" )
# 1336 "lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_regexp_options_rec lexbuf __ocaml_lex_state

;;

